#include <u.h>
#include <libc.h>
#include <bio.h>
#include <auth.h>
#include <fcall.h>
#include <thread.h>
#include <9p.h>

enum
{
	EESDOpInit				= 0,
	EESDOpLock				= 1,
	EESDOpUnLock			= 2,
	EESDOpStreamPlay		= 3,
	EESDOpStreamMonitor		= 5,
	EESDOpServerAllInfo		= 17,
	EESDOpStreamPan			= 20,
};

enum
{
	EESDFormat8Bit			= 0x0000,
	EESDFormat16Bit			= 0x0001,
	EESDFormatMono			= 0x0010,
	EESDFormatStereo		= 0x0020,
	EESDFormatStream		= 0x0000,
	EESDFormatSample		= 0x0100,
	EESDFormatADPCM			= 0x0200,
	EESDFormatStreamPlay	= 0x1000,
	EESDFormatStreamMonitor	= 0x0000,
	EESDFormatStreamRecord	= 0x2000,
	EESDFormatSampleStop	= 0x0000,
	EESDFormatSampleLoop	= 0x2000,
};

enum
{
	EPlayerMax		= 100,
	ENameMax		= 128,
};

typedef struct ESDServerInfo ESDServerInfo;
typedef struct ESDStream ESDStream;

struct ESDServerInfo
{
	int		version;
	int		rate;
	int		format;

	int		leftvolume;
	int		rightvolume;

	int		streamids[EPlayerMax];
};

struct ESDStream
{
	int		fd;
};

enum 
{
	Qdir,
	Qaudio,
	Qaudioin,
	Qvolume,
	Nqid,
};

char*	dialaddr	= nil;
char	esdkey[16];
char*	playername	= "esdaudio";

int readdat(int fd, void *buf, int len)
{
	int c, l;
	l = 0;
	while(l < len){
		c = read(fd, &((uchar*)buf)[l], len-l);
		if(c < 0)
			return 0;
		l += c;
	}
	return l;
}

int writedat(int fd, void *buf, int len)
{
	int c, l, u;
	l = 0;
	while(l < len){
		u = len - l;
		u = u > 1024 ? 1024 : u;
		c = write(fd, &((uchar*)buf)[l], u);
		if(c < 0)
			return 0;
		l += c;
	}
	return l;
}

long
writeint(int fd, int v)
{
	char b[4];

	b[0] = (v & 0x000000ff) >> 0;
	b[1] = (v & 0x0000ff00) >> 8;
	b[2] = (v & 0x00ff0000) >>16;
	b[3] = (v & 0xff000000) >>24;

	return (writedat(fd, b, 4)==4);
}

long
readint(int fd, int *v)
{
	int l;
	char b[4];

	if (l = (readdat(fd, b, 4) == 4))
		*v =	((b[0]<<0)	&	0x000000ff) | 
			((b[1]<<8)	&	0x0000ff00) |
			((b[2]<<16) &	0x00ff0000) |
			((b[3]<<24) &	0xff000000);
	else
		*v = 0;

	return l;
}

int
esdopen(void)
{
	int fd;
	int ok;

	fd = dial(dialaddr, 0, 0, 0);
	if(fd < 0) {
		return -1;
	}

	if (writedat(fd, esdkey, 16)!=16)
		return -1;

	if (writedat(fd, "NDNE", 4)!=4)
		return -1;

	if(!readint(fd, &ok)) {
		close(fd);
		return -1;
	}

	return fd;
}

ESDServerInfo*
esdgetinfo(int fd, ESDServerInfo *info)
{
	int id;
	int version = 0;
	char name[ENameMax];
	char unused[152];
	int dump, leftvolume, rightvolume;
	int i;

	// opcode
	if (!writeint(fd, EESDOpServerAllInfo))
		return nil;

	// client version
	if (!writeint(fd, version))
		return nil;

	// server version
	if (!readint(fd, &info->version))
		return nil;

	// server sample-rate
	if (!readint(fd, &info->rate))
		return nil;

	// server format
	if (!readint(fd, &info->format))
		return nil;

	info->leftvolume = 0;
	info->rightvolume = 0;
	memset(info->streamids, 0, EPlayerMax*sizeof(int));
	i = 0;
	do {
		// stream id
		if (!readint(fd, &id))
			return nil;

		// player name
		if (!readdat(fd, name, ENameMax))
			return nil;
		name[ENameMax-1] = '\0';

		// player rate (ignore, we use server rate)
		if (!readint(fd, &dump))
			return nil;

		// player leftvolume
		if (!readint(fd, &leftvolume))
			return nil;

		// player rightvolume
		if (!readint(fd, &rightvolume))
			return nil;

		// player format (ignore, we use server format)
		if (!readint(fd, &dump))
			return nil;

		if (id > 0) {
			if (strcmp(name, playername) == 0) {
				if (i == 0) {
					info->leftvolume = leftvolume;
					info->rightvolume = rightvolume;
				}
				if (i < EPlayerMax)
					info->streamids[i++] = id;
			}
		}
	} while(id > 0);

	// skip: we don't really care about samples
	do {
		// sample id
		if (!readint(fd, &id))
			return nil;

		// name
		if (!readdat(fd, unused, ENameMax))
			return nil;

		// rate
		if (!readint(fd, &dump))
			return nil;

		// left
		if (!readint(fd, &dump))
			return nil;

		// right
		if (!readint(fd, &dump))
			return nil;

		// format
		if (!readint(fd, &dump))
			return nil;

		// length
		if (!readint(fd, &dump))
			return nil;
	} while(id > 0);

	return info;
}

int
esdpan(int fd, int left, int right)
{
	int i, ok;
	ESDServerInfo esrvinfo;

	if (!esdgetinfo(fd, &esrvinfo))
		return 0;

	for(i=0; i<EPlayerMax; i++) {
		if (esrvinfo.streamids[i] == 0)
			break;

		// opcode
		if (!writeint(fd, EESDOpStreamPan))
			return 0;

		// stream id
		if (!writeint(fd, esrvinfo.streamids[i]))
			return 0;

		// left volume
		if (!writeint(fd, left))
			return 0;

		// right volume
		if (!writeint(fd, right))
			return 0;

		// result
		if (!readint(fd, &ok))
			return 0;

		if (!ok)
			return 0;
	}

	return 1;
}

int
esdplay(int fd)
{
	ESDServerInfo esrvinfo;
	char buf[ENameMax];

	if (!esdgetinfo(fd, &esrvinfo))
		return 0;

	// opcode
	if (!writeint(fd, EESDOpStreamPlay))
		return 0;

	// format
	if (!writeint(fd, esrvinfo.format|EESDFormatStreamPlay))
		return 0;

	// rate
	if (!writeint(fd, esrvinfo.rate))
		return 0;

	// player name
	memset(buf, 0, ENameMax);
	strcat(buf, playername);
	if (!writedat(fd, buf, ENameMax))
		return 0;
	
	return 1;
}

int
esdrecord(int fd)
{
	ESDServerInfo esrvinfo;
	char buf[ENameMax];

	if (!esdgetinfo(fd, &esrvinfo))
		return 0;

	// opcode
	if (!writeint(fd, EESDOpStreamMonitor))
		return 0;

	// format
	if (!writeint(fd, esrvinfo.format|EESDFormatStreamMonitor))
		return 0;

	// rate
	if (!writeint(fd, esrvinfo.rate))
		return 0;

	// player name
	memset(buf, 0, ENameMax);
	strcat(buf, playername);
	if (!writedat(fd, buf, ENameMax))
		return 0;

	return 1;
}

int
fillstat(ulong qid, Dir *d)
{
	memset(d, 0, sizeof(Dir));

	d->uid = "audio";
	d->gid = "audio";
	d->muid = "";
	d->qid = (Qid){qid, 0, 0};
	d->atime = time(0);

	switch(qid) {
	case Qdir:
		d->name = "/";
		d->qid.type = QTDIR;
		d->mode = DMDIR|0777;
		break;

	case Qaudio:
		d->name = "audio";
		d->mode = 0222;
		break;

	case Qaudioin:
		d->name = "audioin";
		d->mode = 0444;
		break;

	case Qvolume:
		d->name = "volume";
		d->mode = 0666;
		break;
	}

	return 1;
}


int
readtopdir(Fid*, uchar *buf, long off, int cnt, int blen)
{
	int i, m, n;
	long pos;
	Dir d;

	n = 0;
	pos = 0;
	for (i = 1; i < Nqid; i++){
		fillstat(i, &d);
		m = convD2M(&d, &buf[n], blen-n);
		if(off <= pos){
			if(m <= BIT16SZ || m > cnt)
				break;
			n += m;
			cnt -= m;
		}
		pos += m;
	}
	return n;
}

long
readvolume(int fd, char *buf, long off, long len)
{
	ESDServerInfo esrvinfo;
	char s[1024];
	int l;

	if (esdgetinfo(fd, &esrvinfo))
		sprint(s, "audio %d %d\nspeed %d\n", 
			esrvinfo.leftvolume*100/256,
			esrvinfo.rightvolume*100/256,
			esrvinfo.rate);

	l = strlen(s)-off;
	if (l < 0)
		l = 0;

	if (len < l)
		l = len;

	if (l > 0)
		memcpy(buf, &s[off], l);

	return l;
}

long
writevolume(int fd, char *buf, long len)
{
	char *cols[4];
	int ncols;
	int leftvolume, rightvolume;

	leftvolume = 100;
	rightvolume = 100;

	ncols = getfields(buf, cols, 4,1, " ");
	switch(ncols) {
	case 1:
		// volume
		rightvolume = leftvolume = strtol(cols[0], nil, 0);
		break;

	case 2:
		// audio volume
		rightvolume = leftvolume = strtol(cols[1], nil, 0);
		break;

	case 3:
		if ((strcmp(cols[1], "out") == 0) || (strcmp(cols[1], "in")==0)) {
			// audio in volume
			// audio out volume
			rightvolume = leftvolume = strtol(cols[2], nil, 0);
		} else {
			// audio leftvolume rightvolume
			leftvolume = strtol(cols[1], nil, 0);
			rightvolume = strtol(cols[2], nil, 0);
		}
		break;

	case 4:
		// audio in leftvolume rightvolume
		leftvolume = strtol(cols[2], nil, 0);
		rightvolume = strtol(cols[3], nil, 0);
		break;
	}
	
	return (esdpan(fd, leftvolume*256/100, rightvolume*256/100) ? len : 0);
}

static void
fsattach(Req *r)
{
	char *spec;

	spec = r->ifcall.aname;
	if(spec && spec[0]) {
		respond(r, "invalid attach specifier");
		return;
	}

	r->fid->qid = (Qid){Qdir, 0, QTDIR};
	r->ofcall.qid = r->fid->qid;
	respond(r, nil);
}

static void
fsstat(Req *r)
{
	fillstat((ulong)r->fid->qid.path, &r->d);

	r->d.name = estrdup9p(r->d.name);
	r->d.uid = estrdup9p(r->d.uid);
	r->d.gid = estrdup9p(r->d.gid);
	r->d.muid = estrdup9p(r->d.muid);

	respond(r, nil);
}

static char*
fswalk1(Fid *fid, char *name, Qid *qid)
{
	switch((ulong)fid->qid.path) {
	case Qdir:
		if (strcmp(name, "..") == 0) {
			*qid = (Qid){Qdir, 0, QTDIR};
			fid->qid = *qid;
			return nil;
		}

		if (strcmp(name, "audio") == 0) {
			*qid = (Qid){Qaudio, 0, 0};
			fid->qid = *qid;
			return nil;
		}

		if (strcmp(name, "audioin") == 0) {
			*qid = (Qid){Qaudioin, 0, 0};
			fid->qid = *qid;
			return nil;
		}

		if (strcmp(name, "volume") == 0) {
			*qid = (Qid){Qvolume, 0, 0};
			fid->qid = *qid;
			return nil;
		}

		return "file not found";
		
	default:
		return "walk in non-directory";
	}
}


static void
fsopen(Req *r)
{
	ESDServerInfo esrvinf;
	ESDStream *stream;
	int omode;
	Fid *fid;
	ulong path;

	fid = r->fid;
	path = (ulong)fid->qid.path;
	omode = r->ifcall.mode;

	
	switch(path) {
	case Qdir:
		if (omode == OREAD)
			respond(r, nil);
		else
			respond(r, "permission denied");
		return;

	default:
		switch(path) {
		case Qaudio:
			if (omode != OREAD)
				goto permok;
			break;

		case Qaudioin:
			if (omode == OREAD)
				goto permok;
			break;

		case Qvolume:
			goto permok;
			break;
		}

		respond(r, "permission denied");
		return;

permok:
		stream = emalloc9p(sizeof(ESDStream));
		stream->fd = esdopen();

		if (stream->fd < 0) {
			free(stream);
			respond(r, "cant connect to esd");
			return;
		}

		switch(path) {
		case Qaudio:
			if (!esdplay(stream->fd)) {
				close(stream->fd);
				free(stream);
				respond(r, "cant play to esd");
				return;
			}
			break;

		case Qaudioin:
			if ((!esdgetinfo(stream->fd, &esrvinf)) || (!esrvinf.streamids[0])) {
				close(stream->fd);
				free(stream);
				respond(r, "no players active on esd");
				return;
			}

			if (!esdrecord(stream->fd)) {
				close(stream->fd);
				free(stream);
				respond(r, "cant record from esd");
				return;
			}
			break;
		}

		fid->aux = stream;
		respond(r, nil);
		return;
	}
}

static void
fsread(Req *r)
{
	uchar *buf;
	long count;
	vlong offset;
	ESDStream *stream;

	offset = r->ifcall.offset;
	count = r->ifcall.count;
	buf = (uchar*)r->ofcall.data;

	switch((ulong)r->fid->qid.path) {
	case Qdir:
		r->ofcall.count = readtopdir(r->fid, buf, offset, count, count);
		respond(r, nil);
		return;

	case Qaudioin:
		if (stream = r->fid->aux)
			count = readdat(stream->fd, buf, count);
		else
			count = 0;
		break;

	case Qvolume:
		if (stream = r->fid->aux)
			count = readvolume(stream->fd, (char*)buf, offset, count);
		else
			count = 0;
		break;
	}

	r->ofcall.count = count;
	respond(r, nil);
	return;
}

static void
fswrite(Req *r)
{
	ESDStream *stream;
	char *buf;
	vlong offset;
	long count;

	offset = r->ifcall.offset;
	count = r->ifcall.count;
	buf = r->ifcall.data;

	switch((ulong)r->fid->qid.path) {
	case Qaudio:
		if (stream = r->fid->aux) {
			if(!writedat(stream->fd, buf, count)) {
				respond(r, "esd write failed");
				return;
			}
		} else {
			count = 0;
		}

		break;

	case Qvolume:
		if (stream = r->fid->aux)
			count = writevolume(stream->fd, buf, count);
		else
			count = 0;
		break;
	}

	r->ofcall.count = count;
	respond(r, nil);
	return;
}

static void
fsdestroyfid(Fid *fid)
{
	ESDStream *stream;
	switch((ulong)fid->qid.path) {
	case Qaudio:
	case Qaudioin:
	case Qvolume:
		if (stream = fid->aux) {
			fid->aux = nil;
			close(stream->fd);
			free(stream);
		}
		break;
	}
}


Srv fs = {
	.attach=		fsattach,
	.walk1=			fswalk1,
	.open=			fsopen,
	.read=			fsread,
	.write=			fswrite,
	.stat=			fsstat,
	.destroyfid=	fsdestroyfid,
};

void 
readkeyfile(char *keyfile)
{
	int fd;

	fd = open(keyfile, OREAD);
	if (fd < 0) {
		perror("open keyfile");
		exits("keyfile");
	}

	if (readdat(fd, esdkey, 16) != 16) {
		print("%s invalid keyfile size\n", keyfile);
		exits("keyfile");
	}

	close(fd);
}

void
usage(void)
{
	fprint(2, "usage: [-k keyfile] [-m mountpoint] net!server!port\n");
	exits("usage");
}

void
main(int argc, char** argv)
{
	char* mtpt = "/dev";

	ARGBEGIN{
	case 'k':
		readkeyfile(EARGF(usage()));
		break;

	case 'm':
		mtpt = EARGF(usage());
		break;

	default:
		usage();
	}ARGEND

	dialaddr = *argv;
	if (!dialaddr) {
		usage();
	}

	postmountsrv(&fs, nil, mtpt, MBEFORE);
	exits(nil);
}
